# Copyright (C) 2023 Intel Corporation
#
# SPDX-License-Identifier: MIT

DATUMARO_FORMAT_VERSION = "1.0"


class DatumaroPath:
    IMAGES_DIR = "images"
    ANNOTATIONS_DIR = "annotations"
    PCD_DIR = "point_clouds"
    VIDEO_DIR = "videos"
    MASKS_DIR = "masks"

    ANNOTATION_EXT = ".json"
    IMAGE_EXT = ".jpg"
    MASK_EXT = ".png"
